-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/ace

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.ace_pkg.all;
use work.ctrl_pkg.all;

entity dp is
  port
    ( clk         : in  std_logic
    ; i_control 	: in 	dp_op_ty
    ; i_new_ctrl 	: in  en_ty
    ; i_pad_ctrl	: in  padxor_ty
    ; i_oh        : in  onehot_ty
    ; i_is_hash   : in  std_logic
    ; i_dom_sep   : in  dom_sep_ty
    ; i_data      : in  word
    ; o_data      : out word
   );
end entity;

architecture rtl of dp is

  signal sel_input, pad_old,
         en_out, sel_round_step,
         sel_load_perm, en_load,
         lfsr_c_reset, lfsr_c_en          : std_logic;

  signal sel_out : std_logic_vector(1 downto 0);

  signal ace_state, post_input, post_pad  : half_word_state_ty;
  signal pre_round, post_round, post_xor,
         post_step_const, post_linear,
         ace_path                         : word_state_ty;
  signal dsxor                            : half_word;
  signal i_data_vector, o_data_vector,
         osel_data_vector                 : half_word_data;
  signal ctl_const                        : lfsr_c_output;
begin

  u_lfsr_c :  entity work.lfsr_c port map
      (	clk, lfsr_c_en,  lfsr_c_reset, ctl_const );

  i_data_vector <= data_to_vector( i_data );
  o_data        <= vector_to_data( o_data_vector );

  sel_input       <= i_control( sel_input_idx );
  sel_out(1)      <= i_control( sel_out1_idx );
  sel_out(0)      <= i_control( sel_out0_idx );
  en_out          <= i_control( en_out_idx );
  sel_round_step  <= i_control( sel_round_step_idx );
  sel_load_perm   <= i_control( sel_load_perm_idx );
  en_load         <= i_control( en_load_idx );
  lfsr_c_reset    <= i_control( lfsr_c_reset_idx );
  lfsr_c_en       <= i_control( lfsr_c_en_idx );

  ----------------------------------------------------------------------
  -- post input:  do input and domain separator

  post_input( a1_idx ) <=  i_data_vector(0)                 when sel_input = '1'
                     else  i_data_vector(0) xor ace_state( a1_idx );
  post_input( c1_idx ) <=  i_data_vector(1)                 when sel_input = '1'
                     else  i_data_vector(1) xor ace_state( c1_idx );

  post_input_cap : for i in 0 to half_word_max_idx generate
    cap_if: if capacity_idx(i) = '1'  generate
      post_input(i) <= ace_state(i);
    end generate  cap_if;
  end generate  post_input_cap ;

  dsxor( 0 to half_word_sz - 3 ) <= ( others => '0' );
  dsxor( half_word_sz - 2)       <= i_dom_sep(1);
  dsxor( half_word_sz - 1)       <= i_dom_sep(0);

  post_input( e0_idx ) <= dsxor xor ace_state( e0_idx );

  ----------------------------------------------------------------------
  -- post pad: padding, select new or internal path

  post_pad_cap : for i in 0 to half_word_max_idx generate
    cap_if: if capacity_idx(i) = '1'  generate
      post_pad(i) <= post_input(i);
    end generate  cap_if;
  end generate  post_pad_cap ;


  post_pad( e0_idx ) <=   post_input( e0_idx ) when i_new_ctrl(0) = '1' else
                          ace_state( e0_idx );
  post_pad( a1_idx )( 0 to 7 ) <=
          post_input( a1_idx )( 0 to 7 ) when i_new_ctrl(0) = '1' else
          ace_state ( a1_idx )( 0 to 7 );

  padbits_A : for i in 1 to 3 generate
    post_pad( a1_idx )( 8*i )  <=
          post_input( a1_idx )( 8*i ) when i_new_ctrl(i) = '1' else
          ace_state ( a1_idx )( 8*i ) xor  i_pad_ctrl(i-1);
  end generate  padbits_A ;

  nonpadbits_A : for i in 1 to 3 generate
    otherbits_A : for j in 1 to 7 generate
    post_pad( a1_idx )( 8*i + j )  <=
          post_input( a1_idx )( 8*i + j ) when i_new_ctrl(i) = '1' else
          ace_state ( a1_idx )( 8*i + j );
    end generate otherbits_A;
  end generate  nonpadbits_A;

  padbits_C : for i in 0 to 3 generate
      post_pad( c1_idx )( 8*i )  <=
          post_input( c1_idx )( 8*i ) when i_new_ctrl(4+i) = '1' else
          ace_state ( c1_idx )( 8*i ) xor  i_pad_ctrl(4+i-1);
  end generate  padbits_C ;

  nonpadbits_C : for i in 0 to 3 generate
    otherbits_C : for j in 1 to 7 generate
      post_pad( c1_idx )( 8*i + j )  <=
          post_input( c1_idx )( 8*i + j ) when i_new_ctrl(4+i) = '1' else
          ace_state ( c1_idx )( 8*i + j );
    end generate otherbits_C;
  end generate  nonpadbits_C;

  ----------------------------------------------------------------------
  --  output control

  osel_data_vector(0) <= i_data_vector(0) xor ace_state( a1_idx )
                                              when sel_out = "00"  else --enc/dec
                         ace_state ( a1_idx ) when sel_out = "01"  else --hash
                         ace_state ( a1_idx ) when sel_out = "10"  else --T0
                         ace_state ( c1_idx);                           --T1

  o_data_vector(0) <= osel_data_vector(0) when en_out = '1' else (others => '0');

  osel_data_vector(1) <= i_data_vector(1) xor ace_state( c1_idx )
                            when sel_out = "00"  else --enc/dec
                         ace_state ( c1_idx ) when sel_out = "01"  else --hash
                         ace_state ( a0_idx ) when sel_out = "10"  else --T0
                         ace_state ( c0_idx);                           --T1

  o_data_vector(1) <= osel_data_vector(1) when en_out = '1' else (others => '0');

  ---------------------------------------------------------------------
  -- sb 64 ==> post round

  pre_round <= half_words_to_words( post_pad );

  a_sb_64 : entity work.sb_64 port map
      ( pre_round( a_idx ), ctl_const( lfsr_c_sz + 2 )   --rc0
      , post_round( a_idx ) );

  post_round( b_idx ) <= pre_round( b_idx );

  c_sb_64 : entity work.sb_64 port map
      ( pre_round( c_idx ), ctl_const( lfsr_c_sz + 1 )   --rc1
      , o_state => post_round( c_idx ) );

  post_round( d_idx ) <= pre_round( d_idx );

  e_sb_64 :  entity work.sb_64 port map
      ( pre_round( e_idx ),ctl_const( lfsr_c_sz )       -- rc2
      , post_round( e_idx ) );

  ----------------------------------------------------------------------
  -- XORs to the left ==> post xor

  post_xor( a_idx ) <= post_round( a_idx );
  post_xor( c_idx ) <= post_round( c_idx );

  post_xor( b_idx ) <= post_round( b_idx ) xor post_round( c_idx );
  post_xor( d_idx ) <= post_round( d_idx ) xor post_round( e_idx );
  post_xor( e_idx ) <= post_round( e_idx ) xor post_round( a_idx );

  ----------------------------------------------------------------------
  -- XOR with step constant ==> post step const

  post_step_const( a_idx )           <= post_xor( a_idx );
  post_step_const( c_idx )           <= post_xor( c_idx );

  post_step_const( b_idx)( 0 to 55)  <= not post_xor( b_idx )( 0 to 55  );
  post_step_const( b_idx)( 56 to 63) <= post_xor(     b_idx )( 56 to 63 ) xor
                                        ctl_const( 2 to lfsr_c_sz + 2 ); -- sc0

  post_step_const( d_idx)( 0 to 55)  <= not post_xor( d_idx )( 0 to 55  );
  post_step_const( d_idx)( 56 to 63) <= post_xor(     d_idx )( 56 to 63 ) xor
                                        ctl_const( 1 to lfsr_c_sz + 1 ); -- sc1

  post_step_const( e_idx)( 0 to 55)  <= not post_xor( e_idx )( 0 to 55  );
  post_step_const( e_idx)( 56 to 63) <= post_xor(     e_idx )( 56 to 63 ) xor
                                        ctl_const( 0 to lfsr_c_sz );     --sc2

  ------------------------------------------------------------
  -- post liear layer pi = (3,2,0,4,1) ==> post linear

  post_linear( a_idx ) <= post_step_const( d_idx );
  post_linear( b_idx ) <= post_step_const( c_idx );
  post_linear( c_idx ) <= post_step_const( a_idx );
  post_linear( d_idx ) <= post_step_const( e_idx );
  post_linear( e_idx ) <= post_step_const( b_idx );

  ------------------------------------------------------------
  -- update state

  ace_path( a_idx ) <= post_linear( a_idx ) when sel_round_step = '1'  else
                                                            post_round( a_idx );
  ace_path( b_idx ) <= post_linear( b_idx ) when sel_round_step = '1'  else
                                                            post_round( b_idx );
  ace_path( c_idx ) <= post_linear( c_idx ) when sel_round_step = '1'  else
                                                            post_round( c_idx );
  ace_path( d_idx ) <= post_linear( d_idx ) when sel_round_step = '1'  else
                                                            post_round( d_idx );
  ace_path( e_idx ) <= post_linear( e_idx ) when sel_round_step = '1'  else
                                                            post_round( e_idx );

  process
  begin

    wait until rising_edge(clk);
    -- ACE_AE
    if i_is_hash = '0' then
      if sel_load_perm = '1' then
        ace_state <= words_to_half_words( ace_path ); -- every clk cycle
      else
        if en_load = '1'   then     -- load
          if i_oh(0) = '1' then
            ace_state( a0_idx ) <= i_data_vector(1);  -- k0
            ace_state( a1_idx ) <= i_data_vector(0);
          end if;
          if i_oh(1) = '1' then
            ace_state( c0_idx ) <= i_data_vector(1);  -- k1
            ace_state( c1_idx ) <= i_data_vector(0);
          end if;
          if i_oh(2) = '1' then
            ace_state( b0_idx ) <= i_data_vector(1);  -- n0
            ace_state( b1_idx ) <= i_data_vector(0);
          end if;
          if i_oh(3) = '1' then
            ace_state( e0_idx ) <= i_data_vector(1);  -- n1
            ace_state( e1_idx ) <= i_data_vector(0);
            ace_state( d0_idx ) <= ( others => '0' );
            ace_state( d1_idx ) <= ( others => '0' );
          end if;
        end if;
      end if;
    -- ACE_H
    else
      if sel_load_perm = '1' then
        ace_state <= words_to_half_words( ace_path );
      else
        if en_load = '1' then                        -- load
          ace_state( b0_idx ) <= i_data_vector(1);
          ace_state( b1_idx ) <= i_data_vector(0);
          ace_state( a0_idx ) <= ( others => '0' );
          ace_state( a1_idx ) <= ( others => '0' );
          ace_state( c0_idx ) <= ( others => '0' );
          ace_state( c1_idx ) <= ( others => '0' );
          ace_state( d0_idx ) <= ( others => '0' );
          ace_state( d1_idx ) <= ( others => '0' );
          ace_state( e0_idx ) <= ( others => '0' );
          ace_state( e1_idx ) <= ( others => '0' );
        end if;
      end if;
    end if;

  end process;

end architecture;
