-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2020 VHDL modifications:  Mark D. Aagaard, Nusa Zidaric

-- Mark D. Aagaard
-- Riham AlTawy
-- Guang Gong
-- Kalikinkar Mandal
-- Raghvendra Rohit
-- Marat Sattarov
-- Nusa Zidaric
-- http://uwaterloo.ca/communications-security-lab/lwc/ace


-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.

---------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.all;
USE STD.textio.all;
USE ieee.std_logic_textio.all;

use work.util_unsynth.all;
use work.util.all;
use work.tb_ace_unsynth.all;
use work.ace_pkg.all;
use work.ctrl_pkg.all;
use work.tb_pkg.all;

ENTITY tb_ace IS
  generic ( period : real   := 10.0
          ; EDH    : std_logic_vector(0 to 2) := "001"--"110" -- Encryption | Decryption | Hash
          ; S      : std_logic_vector(0 to 1) := "00" -- stall flag: 00 - none, 01 - stall enc, 10 - stall dec, 11 stall other
         );
END tb_ace;

ARCHITECTURE main OF tb_ace IS

  file simo_file      : text open write_mode is simo_file_path;
  file stim_file      : text;

  constant clk_period   : time  := period * 1 ns;
  constant hold         : time  := clk_period / 5;
  constant after_reset  : time  := 0 * clk_period;  -- delay between reset and loading
  constant load_delay   : time  := 0 * clk_period;  -- delay between K, N data
  constant init_delay   : time  := 0 * clk_period;  -- delay between load permutation and initialization
  constant procad_delay : time  := 0 * clk_period;  -- AD data delay
  constant stall_short_delay  : time  := 3   * clk_period;  -- stall short delay
  constant stall_long_delay   : time  := 300 * clk_period;  -- stall long delay
  constant stall_period       : time  := 5   * clk_period;
  constant enc_delay    : time  := 10 * clk_period;  -- encryption data delay
  constant dec_delay    : time  := 10 * clk_period;  -- decryption data delay
  constant mode_delay   : time  := 50 * clk_period;  -- delay between two modes



  signal clk, reset : std_logic := '0';
  signal stall : boolean := false;

  signal tag_data : std_logic_vector(0 to 127);
  signal hash_data : std_logic_vector(0 to 255);
  signal tag_valid, hash_valid  : std_logic := '0';
  type tb_state_ty is (tbLoad, tbInit, tbProcAD, tbEncrypt, tbDecrypt, tbFinal, tbTag, tbAbsorb, tbSqueeze, tbNull);
  signal tb_state : tb_state_ty;

  signal pre_data, post_data    : word;
  signal pre_valid, pre_ready   : std_logic;
  signal post_valid, post_ready : std_logic;
  signal pre_dom_sep     : dom_sep_ty;
  signal pre_dom_sep_buf : dom_sep_ty;
  signal pre_mode        : mode_ty;
  signal pre_mode_buf    : mode_ty;
  signal pre_padding     : full_pad_ty;
  signal bits_unpadded   : integer;

BEGIN

  pre_dom_sep_proc : process(pre_dom_sep_buf)
  begin
    pre_dom_sep <= pre_dom_sep_buf after hold;
  end process;

  pre_mode_proc : process(pre_mode_buf)
  begin
    pre_mode <= pre_mode_buf after hold;
  end process;

  post_reading_proc : process -- reading the tb/sim output , writing the simo file
    variable msg    : line;
    variable done, skip_write  : boolean;
    variable post_data_var : std_logic_vector(0 to 63);
  begin

    skip_write := false;
    done := false;
    while not(done) loop
      wait until rising_edge(clk);
      done := (post_valid = '1' or tag_valid = '1' or hash_valid = '1' or (pre_valid = '1' and pre_ready='1'));
    end loop;

    post_data_var := post_data;

	    if tb_state = tbEncrypt then
			 if 	bits_unpadded >= 0  then
			      write(  msg, "C " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));
			 else
	       		skip_write := true;
	      end if;
	    elsif tb_state = tbDecrypt then
		    if 	bits_unpadded >= 0 then
	     			write(  msg, "P " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));
			 else
	       		skip_write := true;
	       end if;

	    elsif tb_state = tbTag then
	      if tag_valid = '1' then
	        write(  msg, "T " & to_hex_string_normal( tag_data ));
	      else
	        skip_write := true;
	      end if;

		elsif tb_state = tbSqueeze then
	      if hash_valid = '1' then
	        write(  msg, "H " & to_hex_string_normal( hash_data ));
	      else
	        skip_write := true;
	      end if;

	    else
	      skip_write := true;
	    end if;

    if not(skip_write) then writeline( simo_file, msg); end if;

  end process;

  post_stalling_proc : process

  begin

      if S = "01" then
      	if   tb_state = tbEncrypt then
	 		 	stall <= true;
				wait for stall_short_delay;
	 		end if;
	 		stall <= false;
	 		wait for stall_period;
	 	elsif S = "10" then
      	if   tb_state = tbDecrypt then
	 		 	stall <= true;
				wait for stall_short_delay;
	 		end if;
	 		stall <= false;
	 		wait for stall_period;
	 	elsif S = "11" then
      	if tb_state = tbInit or tb_state = tbProcAD or tb_state = tbFinal or  tb_state = tbTag  or tb_state = tbAbsorb or  tb_state = tbSqueeze then
	 		 	stall <= true;
				--wait for stall_long_delay;
				wait for stall_short_delay;
	 		end if;
	 		stall <= false;
	 		wait for stall_period;

 		else
 		   stall <= false;
 		   wait for stall_period;
  		end if;

  end process;

  post_ready <= '0' when stall else '1';
  clk <= not(clk) after clk_period / 2;


   -- Instantiate the Unit Under Test (UUT)
  uut: entity work.ace port map
    (  clk       => clk
    ,  reset     => reset
    ,  pre_mode    => pre_mode
    ,  pre_dom_sep => pre_dom_sep
    ,  pre_padding => pre_padding
    ,  pre_valid   => pre_valid
    ,  post_ready  => post_ready
    ,  pre_data    => pre_data
    ,  post_data    => post_data
    ,  post_valid   => post_valid
    ,  pre_ready    => pre_ready
     );



  stim_proc: process

    variable target_stim					: stim_ty;
    variable key_rd, nonce_rd 			: std_logic_vector(0 to 127);
    variable k0, k1, n0, n1, data, iv 	: std_logic_vector(0 to 63);


  begin

    if EDH(0) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Encryption --------------------------------
      ------------------------------------------------------------------------------

      tag_data <= (others => 'X');
      pre_valid  <= '0';
      pre_padding  <= (others => '0');

      pre_mode_buf <= "00"; -- new FSM requires this stable
      tb_state <= tbNull;

      -------------------------------- Reset -------------------------------

      wait for clk_period;
      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);

      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      pre_dom_sep_buf <= "00";

      read_key_nonce( key_rd, nonce_rd, stim_file_path );
      k0 := key_rd(0 to word_sz - 1);
      k1 := key_rd(word_sz to key_sz - 1);
      n0 := nonce_rd(0 to word_sz - 1);
      n1 := nonce_rd(word_sz to key_sz - 1);


      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, n0, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, n1, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;

      -------------------------------- LoadPerm ------------------------------
      report( "PERMUTATION AFTER LOAD" );
      wait until rising_edge(clk);
      wait for init_delay;


    --   ---------------------------------- Init --------------------------------
      report( "INITIALIZATION PHASE" );
      tb_state <= tbInit;

      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

    --
    --   ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;

      pre_dom_sep_buf <= "01";

		  target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

    --
    --   ---------------------------------- Encrypt ------------------------------
      report( "ENCRYPTION PHASE" );
      tb_state <= tbEncrypt;

      pre_dom_sep_buf <= "10";


		  target_stim := PLAINTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, enc_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

    --   ---------------------------------- Final --------------------------------
      report( "FINALIZATION PHASE" );
      tb_state <= tbFinal;

      pre_dom_sep_buf <= "00";

	  	target_stim := KEY_STIM;
      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

    --
    --   ---------------------------------- Tag ----------------------------------
      report( "TAG PHASE" );
      tb_state <= tbTag;

      for i in 0 to 1 loop
         wait until post_valid='1' and post_ready ='1' and rising_edge(clk);
       	tag_data(i*64 to i*64 + 63) <= post_data;
        	report(    "tag vector was read for i=" &integer'image(i) &" : " & to_hex_string_normal( post_data )       );
      end loop;

      tag_valid <= '1';

      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';
      wait for mode_delay;

    end if;

    if EDH(1) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Decryption --------------------------------
      ------------------------------------------------------------------------------

      tag_data <= (others => 'X');
      pre_valid  <= '0';
      pre_padding  <= (others => '0');

      pre_mode_buf <= "01"; -- new FSM requires this stable
      tb_state <= tbNull;

      -------------------------------- Reset -------------------------------

      wait for clk_period;
      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);

      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      pre_dom_sep_buf <= "00";

      read_key_nonce( key_rd, nonce_rd, stim_file_path );
      k0 := key_rd(0 to word_sz - 1);
      k1 := key_rd(word_sz to key_sz - 1);
      n0 := nonce_rd(0 to word_sz - 1);
      n1 := nonce_rd(word_sz to key_sz - 1);


      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, n0, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;
      drive_data( clk, n1, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;

      -------------------------------- LoadPerm ------------------------------
      report( "PERMUTATION AFTER LOAD" );
      wait until rising_edge(clk);
      wait for init_delay;


    --   ---------------------------------- Init --------------------------------
      report( "INITIALIZATION PHASE" );
      tb_state <= tbInit;

      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);
      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

    --
    --   ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;

      pre_dom_sep_buf <= "01";

		  target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

    --
    --   ---------------------------------- Decrypt ------------------------------
      report( "DECRYPTION PHASE" );
      tb_state <= tbDecrypt;

      pre_dom_sep_buf <= "10";


	  	target_stim := CIPHERTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, enc_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

    --   ---------------------------------- Final --------------------------------
      report( "FINALIZATION PHASE" );
      tb_state <= tbFinal;

      pre_dom_sep_buf <= "00";

		  target_stim := KEY_STIM;
      drive_data( clk, k0, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

      drive_data( clk, k1, hold, pre_ready, pre_data, pre_valid );
      wait until rising_edge(clk);

    --
    --   ---------------------------------- Tag ----------------------------------
      report( "TAG PHASE" );
      tb_state <= tbTag;

      for i in 0 to 1 loop
         wait until post_valid='1' and post_ready ='1' and rising_edge(clk);
       	tag_data(i*64 to i*64 + 63) <= post_data;
        	report(    "tag vector was read for i=" &integer'image(i) &" : " & to_hex_string_normal( post_data )       );
      end loop;

      tag_valid <= '1';

      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';
      wait for mode_delay;
    --
     end if;

  if EDH(2) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Hash --------------------------------
      ------------------------------------------------------------------------------

      tag_data <= (others => 'X');
      pre_valid  <= '0';
      pre_padding  <= (others => '0');

 		  pre_mode_buf <= "10";   -- new FSM requires this stable
      tb_state <= tbNull;

      -------------------------------- Reset -------------------------------

      wait for clk_period;
      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);

      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      pre_dom_sep_buf <= "00";

      iv := x"8040400000000000";
      drive_data( clk, iv, hold, pre_ready, pre_data, pre_valid );        wait for load_delay;


      -------------------------------- LoadPerm ------------------------------
      report( "PERMUTATION AFTER LOAD" );
      wait until rising_edge(clk);
      wait for init_delay;

     --------------------------------- Absorb -------------------------------
      report( "HASH ABSORB" );
      tb_state <= tbAbsorb;
      pre_mode_buf <= "10";

      target_stim := PLAINTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready, pre_data, pre_valid, pre_padding,  bits_unpadded);

    --
    --   ---------------------------------- Squeeze----------------------------------
      report( "HASH SQUEEZE" );
      tb_state <= tbSqueeze;
      pre_mode_buf <= "11";
      -- dont care for pre_data, but need input because pre_mode requires pre_valid signal!
      pre_valid  <= '1';


      wait until pre_ready = '1' and post_valid='1' and post_ready ='1' and rising_edge(clk);
      hash_data(0 to 63) <= post_data;
      pre_valid  <= '0';
      report(    "hash vector was read for i=0 : " & to_hex_string_normal( post_data )       );

      for i in 1 to 3 loop

         wait until post_valid='1' and post_ready ='1' and rising_edge(clk);
       	hash_data(i*64 to i*64 + 63) <= post_data;
        	report(    "hash vector was read for i=" &integer'image(i) &" : " & to_hex_string_normal( post_data )       );
      end loop;

      hash_valid <= '1';

      wait until rising_edge(clk); wait for hold;
      report("hash is: " & to_hex_string_normal( hash_data ) );
      hash_valid <= '0';
      wait for mode_delay;
    --
     end if;

      assert false
      report ("SIMULATION IS FINISHED")
      severity failure;

  end process;

end;
